use crate::vdf::Dictionary;

pub struct NonSteamGame {
    /// UNIX timestamp of the last time the game was played
    pub last_play_time: u32,
    /// Location of the shortcut (e.g., .desktop file on Linux) the game entry was generated from
    pub shortcut_path: String,
    /// Executable path for the game
    pub executable: String,
    /// Whether to allow Steam overlay over the game
    pub allow_overlay: bool,
    /// Whether the game is hidden from the Steam library(?)
    pub is_hidden: bool,
    /// Whether the game supports OpenVR
    pub open_vr: bool,
    /// Unknown
    pub devkit_game_id: String,
    /// App name as it appears in Steam
    pub app_name: String,
    /// Starting directory for the game. Default is "./", for the current working directory.
    pub start_dir: String,
    /// What options to launch the game with. Used with `executable`.
    pub launch_options: String,
    /// The app ID of the game. Default is to use a randomly-generated integer, to avoid conflict with other installed games.
    pub app_id: u32,
    /// Unknown. Possibly means this was created by the SteamOS Devkit Client?
    pub devkit: bool,
    /// If the game is a Flatpak image, the Flatpak app ID of the game
    pub flatpak_app_id: String,
    /// If the game has a launcher, allows a Steam controller to use the desktop config to interact with it
    pub allow_desktop_config: bool,
    /// List of tags used for the game
    pub tags: Vec<String>,
    /// Path to the icon for the game. Both .ico and .png seem to be supported
    pub icon: String,
    /// Unknown. More Devkit stuff
    pub devkit_override_app_id: bool,
}

impl NonSteamGame {
    pub fn new<S: AsRef<str> + Default>(
        name: S,
        executable: S,
        launch_options: Option<S>,
        icon: Option<S>,
    ) -> Self {
        let app_id = {
            let input = format!("{}{}", executable.as_ref(), name.as_ref());
            crc32fast::hash(input.as_bytes()) | 0x80000000
        };
        NonSteamGame {
            app_name: name.as_ref().to_owned(),
            executable: executable.as_ref().to_owned(),
            app_id,
            launch_options: launch_options.unwrap_or_default().as_ref().to_owned(),
            icon: icon.unwrap_or_default().as_ref().to_owned(),
            ..Default::default()
        }
    }
}

macro_rules! append_kv {
    ($d:expr, $k:literal, $v:expr) => {
        $d.insert(String::from($k), crate::vdf::VdfValue::from($v))
    };
}

impl Default for NonSteamGame {
    fn default() -> Self {
        NonSteamGame {
            last_play_time: 0,
            shortcut_path: String::new(),
            executable: String::new(),
            allow_overlay: true,
            is_hidden: false,
            open_vr: false,
            devkit_game_id: String::new(),
            app_name: String::new(),
            start_dir: String::from("./"),
            launch_options: String::new(),
            app_id: 0,
            devkit: false,
            flatpak_app_id: String::new(),
            allow_desktop_config: true,
            tags: Vec::new(),
            icon: String::new(),
            devkit_override_app_id: false,
        }
    }
}

impl From<NonSteamGame> for Dictionary {
    fn from(value: NonSteamGame) -> Self {
        let mut dict = Dictionary::new();

        append_kv!(dict, "LastPlayTime", value.last_play_time);
        // TODO: .desktop path is stored here; why does steam use this?
        append_kv!(dict, "ShortcutPath", &value.shortcut_path);
        // quoted string
        append_kv!(dict, "Exe", &value.executable);
        append_kv!(dict, "AllowOverlay", value.allow_overlay);
        append_kv!(dict, "IsHidden", value.is_hidden);
        append_kv!(dict, "OpenVR", value.open_vr);
        append_kv!(dict, "DevkitGameID", &value.devkit_game_id);
        append_kv!(dict, "AppName", &value.app_name);
        append_kv!(dict, "StartDir", &value.start_dir);
        append_kv!(dict, "LaunchOptions", &value.launch_options);
        append_kv!(dict, "appid", value.app_id);
        append_kv!(dict, "Devkit", value.devkit);
        append_kv!(dict, "FlatpakAppID", &value.flatpak_app_id);
        append_kv!(dict, "AllowDesktopConfig", value.allow_desktop_config);
        append_kv!(dict, "tags", crate::vdf::VdfValue::from(value.tags));
        append_kv!(dict, "icon", &value.icon);
        append_kv!(dict, "DevkitOverrideAppID", value.devkit_override_app_id);

        dict
    }
}

macro_rules! map_to_struct {
    ($dict:expr, $st:tt { $($ent:tt => $key:literal),+ } ) => {
        $st {
            $($ent: $dict[$key].clone().try_into()?),+
        }
    };
}

impl TryFrom<Dictionary> for NonSteamGame {
    type Error = Box<dyn std::error::Error>;
    // TODO: do we care if this panics? maybe we should care
    fn try_from(value: Dictionary) -> Result<Self, Self::Error> {
        Ok(map_to_struct!(value, NonSteamGame {
            last_play_time => "LastPlayTime",
            shortcut_path => "ShortcutPath",
            executable => "Exe",
            allow_overlay => "AllowOverlay",
            is_hidden => "IsHidden",
            open_vr => "OpenVR",
            devkit_game_id => "DevkitGameID",
            app_name => "AppName",
            start_dir => "StartDir",
            launch_options => "LaunchOptions",
            app_id => "appid",
            devkit => "Devkit",
            flatpak_app_id => "FlatpakAppID",
            allow_desktop_config => "AllowDesktopConfig",
            tags => "tags",
            icon => "icon",
            devkit_override_app_id => "DevkitOverrideAppID"
        }))
    }
}
